/*
 * gcc -o sxft sdl.c  -I/usr/include/SDL -I/usr/include/X11R6/Xft  -I/usr/include/freetype2/ -L/usr/X11R6/lib -lX11 -lXft -lSDL
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sdl_xft.h"


/* initialize  sdl_xft */
int SDL_Xft_Init ( SDL_Xft_WMInfo *sdl_xft , SDL_Surface *screen) {

	SDL_SysWMinfo info;

	/* get window information */
	SDL_VERSION(&info.version);
	if (SDL_GetWMInfo(&info) <0 ) {
		fprintf(stderr , "Error : SDL_GetWMInfo\n");
		return 0;
	}

	/*set display and window infomation */
	sdl_xft->d = info.info.x11.display;
	sdl_xft->w = info.info.x11.window;

	if (screen ==  NULL ) {
		fprintf(stderr, "Error : Can not found primary display\n");
		return 0;
	}
	sdl_xft->screen=screen;

	sdl_xft->config=FcConfigCreate();
	sdl_xft->config=FcConfigGetCurrent();

	if (sdl_xft->config == NULL) {
		fprintf(stderr, "Error : FcConfig\n");
		return 0;
	}
	return 1;
}

/* get system font set */
FcChar8 **SDL_Xft_Get_FontList (int *fnum, SDL_Xft_WMInfo *sdl_xft) {

	FcFontSet *f_set;
	FcPattern *pattern;

	FcResult result;
	FcResult cnt_result;

	FcChar8 *f_family;
	FcChar8 **fc=NULL;
	FcChar8 *fc_name=NULL;

	int i=0;
	int fsize=0;

	f_set=FcConfigGetFonts(sdl_xft->config, FcSetSystem);;
	*fnum=f_set->nfont;

	/* count size of font list  */
	for (i=0; i < f_set->nfont; i++) {
		pattern=f_set->fonts[i];
		cnt_result=FcPatternGetString (pattern , FC_FAMILY, 0, &f_family);

		fsize+=strlen(f_family);
	}

	if (f_set->nfont == 0 || fsize == 0 ) {
		/* fonts don't exist */
		return NULL;
	}

	fc=calloc(f_set->nfont, sizeof (FcChar8 *));
	fc_name=(FcChar8 *)malloc(fsize*sizeof (FcChar8 *));
	memset(fc_name, '\0', fsize*sizeof(FcChar8 *));
	
	for (i=0 ; i < f_set->nfont; i++) {
		pattern=f_set->fonts[i];
		result=FcPatternGetString (pattern , FC_FAMILY, 0, &f_family);

		if (result == FcResultMatch) {
			strcpy(fc_name, f_family);
			fc[i]=fc_name;
			fc_name+=f_set->sfont;
		}
	}
	return fc;

}

void SDL_Xft_FontList_Free (FcChar8 **f_list) {
	free(f_list[0]);
	free(f_list);
}


/* creating new sdl_xft_surface */
SDL_Xft_Surface * SDL_Xft_Create_Surface (Uint32 flag, SDL_Xft_WMInfo *sdl_xft, 
									unsigned int width, unsigned int height ) {
	SDL_Xft_Surface *xft_surface;

	SDL_Surface *src;
	Display *d;
	Window w;

	XftDraw *xftdraw;
	FcConfig *config;

	d=sdl_xft->d;
	w=sdl_xft->w;

	/* new memory allocation for SDL_Xft_Surface */
	xft_surface=(SDL_Xft_Surface *) malloc ( sizeof (SDL_Xft_Surface));

	/* create sdl surface */
	src=SDL_CreateRGBSurface(flag , width,  height, 
							sdl_xft->screen->format->BitsPerPixel, 
							sdl_xft->screen->format->Rmask, 
							sdl_xft->screen->format->Gmask, 
							sdl_xft->screen->format->Bmask, 
							sdl_xft->screen->format->Amask);
	xft_surface->surface=src;

	/* create pixmap */
	xft_surface->pixmap=XCreatePixmap(d, w , width, height, DefaultDepth(d,0));
	
	/* set default xftpattern to NULL */
	xft_surface->fcpattern=NULL;

	/* set default value */
	xft_surface->color.colorkey=SDL_XFT_NO_COLORKEY;

	/*create xftdraw */
	xft_surface->xftdraw=XftDrawCreate(d, xft_surface->pixmap, 
							DefaultVisual (d, 0),
							DefaultColormap (d, 0));

	/* set default color */

	/* default background color is white */
	xft_surface->color.bg.red = 0xffff;
	xft_surface->color.bg.green = 0xffff;
	xft_surface->color.bg.blue = 0xffff; 
	xft_surface->color.bg.alpha= 0x00ffff; /* alpha doesn't support */

	/* default foregorund color is black */
	xft_surface->color.fg.red = 0x0000;
	xft_surface->color.fg.green = 0x0000;
	xft_surface->color.fg.blue = 0x0000;
	xft_surface->color.fg.alpha= 0x00ffff; /* alpha doesn't not support */

	xft_surface->extents=NULL;

	return xft_surface;
}

void SDL_Xft_Color_Alloc ( SDL_Xft_Color *color , unsigned int flag, 
						Uint16 r, Uint16 g, Uint16 b) {
	switch(flag) {
		case SDL_XFT_BG_COLOR :
			color->bg.red = r;
			color->bg.green = g;
			color->bg.blue = b;
			color->bg.alpha= 0x00ffff; /* alpha doesn't support */
			break;
		case SDL_XFT_FG_COLOR :
			color->fg.red = r;
			color->fg.green = g;
			color->fg.blue = b;
			color->fg.alpha= 0x00ffff; /* alpha doesn't support */
			break;
		default :
			break;
	}
}

/* set colorkey */
void SDL_Xft_Set_Colorkey (SDL_Xft_Color *color , unsigned int flag) {
	color->colorkey=flag;
}

/* FcPatternAddInteger Wrapper */
int SDL_Xft_Pattern_AddInteger (SDL_Xft_Surface *xft_surface,
								const char *object, int i) {
	return FcPatternAddInteger(xft_surface->fcpattern, object, i);
}

/* FcPatternAddDoubel Wrapper */
int SDL_Xft_Pattern_AddDouble (SDL_Xft_Surface *xft_surface, 
								const char *object, double d) {
	return FcPatternAddDouble (xft_surface->fcpattern, object, d);
}

/* FcPatternAddString Wrapper */
int SDL_Xft_Pattern_AddString (SDL_Xft_Surface *xft_surface, 
								const char *object, const char *s) {
	return FcPatternAddString (xft_surface->fcpattern, object, s);
}

/* FcPatternAddMatrix Wrapper */
int SDL_Xft_Pattern_AddMatrix (SDL_Xft_Surface *xft_surface,
								const char *object, const FcMatrix *s) {
	return FcPatternAddMatrix (xft_surface->fcpattern, object, s);
}

/* FcPatternAddCharSet Wrapper */
int SDL_Xft_Pattern_AddCharSet (SDL_Xft_Surface *xft_surface,
								const char *object, const FcCharSet *c) {
	return FcPatternAddCharSet (xft_surface->fcpattern, object, c);
}

/* FcPatternAddBool Wrapper */
int SDL_Xft_Pattern_AddBool (SDL_Xft_Surface *xft_surface,
							const char *object, FcBool b) {
	return FcPatternAddBool (xft_surface->fcpattern, object, b);
}

/* FcPatternGetInteger Wrapper */
unsigned int SDL_Xft_Pattern_GetInteger(SDL_Xft_Surface *xft_surface,
										const char *object, int *i) {
	FcResult result;

	result=FcPatternGetInteger(xft_surface->fcpattern, object, 0, i);
	if (result == FcResultTypeMismatch) {
		return FC_PATTERN_GET_FAIL;
	}
	return FC_PATTERN_GET_SUCCESS;
}

/* FcPatternGetDouble Wrapper */
unsigned int SDL_Xft_Pattern_GetDouble(SDL_Xft_Surface *xft_surface,
										const char *object, double *d) {
	FcResult result;

	result=FcPatternGetDouble(xft_surface->fcpattern, object, 0, d);
	if (result == FcResultTypeMismatch) {
		return FC_PATTERN_GET_FAIL;
	}
	return FC_PATTERN_GET_SUCCESS;
}

/* FcPatternGetString Wrapper */
unsigned int SDL_Xft_Pattern_GetString(SDL_Xft_Surface *xft_surface,
										const char *object, FcChar8 **consts) {
	FcResult result;
	result=FcPatternGetString(xft_surface->fcpattern, object, 0, consts);
	if (result == FcResultTypeMismatch) {
		return FC_PATTERN_GET_FAIL;
	}
	return FC_PATTERN_GET_SUCCESS;
}

/* FcPatternGetMatrix Wrapper */
unsigned int SDL_Xft_Pattern_GetMatrix(SDL_Xft_Surface *xft_surface,
										const char *object, FcMatrix **s) {
	FcResult result;

	result=FcPatternGetMatrix(xft_surface->fcpattern, object, 0, s);
	if (result == FcResultTypeMismatch) {
		return FC_PATTERN_GET_FAIL;
	}
	return FC_PATTERN_GET_SUCCESS;
}

/* FcPatternGetCharSet Wrapper */
unsigned int SDL_Xft_Pattern_GetCharSet(SDL_Xft_Surface *xft_surface,
										const char *object, FcCharSet **c) {
	FcResult result;

	result=FcPatternGetCharSet(xft_surface->fcpattern, object, 0, c);
	if (result == FcResultTypeMismatch) {
		return FC_PATTERN_GET_FAIL;
	}
	return FC_PATTERN_GET_SUCCESS;
}

/* FcPatternGetBool Wrapper */
unsigned int SDL_Xft_Pattern_GetBool(SDL_Xft_Surface *xft_surface,
										const char *object, FcBool *b) {
	FcResult result;

	result=FcPatternGetBool(xft_surface->fcpattern, object, 0, b);
	if (result == FcResultTypeMismatch) {
		return FC_PATTERN_GET_FAIL;
	}
	return FC_PATTERN_GET_SUCCESS;
}

/* Make New FcPattern as FcPatternBuild Wrapper */
void SDL_Xft_Pattern_Build (SDL_Xft_Surface *xft_surface, ...) {

	va_list va;
	FcPattern *set;

	if (xft_surface->fcpattern != NULL) {
		FcPatternDestroy(xft_surface->fcpattern);
	}

	va_start(va, xft_surface);
	set=FcPatternVaBuild(0, va);
	va_end(va);

	xft_surface->fcpattern=FcPatternDuplicate(set);
	FcPatternDestroy(set);
}

void SDL_Xft_Pattern_Create ( SDL_Xft_Surface *xft_surface) {
	xft_surface->fcpattern=FcPatternCreate();
}

/* replace to  matched pattern */
void SDL_Xft_FontMatch(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface * xft_surface) {
	FcPattern *dup_fc ,*match, *tmp_fc;
	FcResult result;

	dup_fc=FcPatternDuplicate(xft_surface->fcpattern);
	FcDefaultSubstitute(dup_fc);
	match=FcFontMatch (sdl_xft->config, dup_fc, &result);
	if (result != FcResultNoMatch) {
		FcPatternDestroy(xft_surface->fcpattern);
		xft_surface->fcpattern=FcPatternDuplicate(match);
	}
	FcPatternDestroy(match);
	FcPatternDestroy(dup_fc);
}

/* XftTextExtemts8 Wrapper */
void SDL_Xft_TextExtents8 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
							FcChar8 *msg) {

	XftFont *xftfont=NULL;
	XftPattern *patterntmp;

	Display *d;

	int s_len=0;

	d=sdl_xft->d;

	if (xft_surface->extents != NULL ) {
		free(xft_surface->extents);
	}
	xft_surface->extents = ( XGlyphInfo *) malloc ( sizeof (XGlyphInfo));
	patterntmp=FcPatternDuplicate(xft_surface->fcpattern);

	while(msg[s_len] != (FcChar8)0) {
		s_len++;
	}
	xftfont=XftFontOpenPattern (d, patterntmp);
	/* this is buggy */
	if(xftfont== NULL)  {
		fprintf(stdout, "No Error !\n");
	}
	XftTextExtents8 (d, xftfont, msg, s_len, xft_surface->extents);
	XftFontClose(d, xftfont);
}

/* XftTextExtemts16 Wrapper */
void SDL_Xft_TextExtents16 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
							FcChar16 *msg) {

	XftFont *xftfont=NULL;
	XftPattern *patterntmp;

	Display *d;

	int s_len=0;

	d=sdl_xft->d;

	if (xft_surface->extents != NULL ) {
		free(xft_surface->extents);
	}
	xft_surface->extents = ( XGlyphInfo *) malloc ( sizeof (XGlyphInfo));
	patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

	while(msg[s_len] != (FcChar16)0) {
		s_len++;
	}
	xftfont=XftFontOpenPattern (d, patterntmp);

	XftTextExtents16 (d, xftfont, msg, s_len, xft_surface->extents);

	XftFontClose(d, xftfont);
}

/* XftTextExtemts32 Wrapper */
void SDL_Xft_TextExtents32 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
							FcChar32 *msg) {

	XftFont *xftfont=NULL;
	XftPattern *patterntmp;

	Display *d;

	int s_len=0;

	d=sdl_xft->d;

	if (xft_surface->extents != NULL ) {
		free(xft_surface->extents);
	}
	xft_surface->extents = ( XGlyphInfo *) malloc ( sizeof (XGlyphInfo));
	patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

	while(msg[s_len] != (FcChar32)0) {
		s_len++;
	}
	xftfont=XftFontOpenPattern (d, patterntmp);

	XftTextExtents32 (d, xftfont, msg, s_len, xft_surface->extents);

	XftFontClose(d, xftfont);
}

/* XftTextExtemts32 Wrapper */
void SDL_Xft_TextExtentsUtf8 (SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
							FcChar8 *msg) {

	XftFont *xftfont=NULL;
	XftPattern *patterntmp;

	Display *d;

	int s_len=0;

	d=sdl_xft->d;

	if (xft_surface->extents != NULL ) {
		free(xft_surface->extents);
	}
	xft_surface->extents = ( XGlyphInfo *) malloc ( sizeof (XGlyphInfo));
	patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

	while(msg[s_len] != (FcChar8)0) {
		s_len++;
	}
	xftfont=XftFontOpenPattern (d, patterntmp);

	XftTextExtentsUtf8 (d, xftfont, msg, s_len, xft_surface->extents);

	XftFontClose(d, xftfont);
}

/* Draw Backcgound Pixmap */
void SDL_Xft_Draw_BackGround (SDL_Xft_WMInfo *sdl_xft, 
						SDL_Xft_Surface *xft_surface,
						unsigned int x, unsigned int y, 
						unsigned int width , unsigned int height) {
	XRenderColor colortmp;
	XftColor bg;

	Display *d;

	d=sdl_xft->d;
	colortmp=xft_surface->color.bg;

	XftColorAllocValue(d, DefaultVisual(d, 0), 
					DefaultColormap(d,0), &colortmp, &bg);

	XftDrawRect(xft_surface->xftdraw, &bg, x, y, width, height);
}

/* draw 8bit string into pixmap of sdl_xft_surface  */
void SDL_Xft_Draw_String8(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
						unsigned int x, unsigned int y, FcChar8 *msg) {
	XftFont *xftfont=NULL;
	XRenderColor colortmp;
	XftColor fg;
	XftPattern *patterntmp;

	Display *d;

	int s_len=0;

	d=sdl_xft->d;
	colortmp=xft_surface->color.fg;

	patterntmp=XftPatternDuplicate(xft_surface->fcpattern);

	while (msg[s_len] != (FcChar8)0) {
		s_len++;
	}

	xftfont=XftFontOpenPattern (d, patterntmp);

	XftColorAllocValue(d, DefaultVisual(d, 0), 
					DefaultColormap(d,0), &colortmp, &fg);

	XftDrawString8(xft_surface->xftdraw, &fg, xftfont, x, y,
					(FcChar8 * ) msg, s_len);
	XftFontClose(d, xftfont);
}

/* draw 16bit string into pixmap of sdl_xft_surface  */
void SDL_Xft_Draw_String16(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
						unsigned int x, unsigned int y, FcChar16 *msg) {
	XftFont *xftfont=NULL;
	XRenderColor colortmp;
	XftColor fg;
	XftPattern *patterntmp;

	Display *d;

	int s_len=0;

	d=sdl_xft->d;
	colortmp=xft_surface->color.fg;

	patterntmp=XftPatternDuplicate(xft_surface->fcpattern);
	
	while (msg[s_len] != (FcChar16)0) {
		s_len++;
	}

	xftfont=XftFontOpenPattern (d, patterntmp);

	XftColorAllocValue(d, DefaultVisual(d, 0), 
					DefaultColormap(d,0), &colortmp, &fg);

	XftDrawString16(xft_surface->xftdraw, &fg, xftfont, x, y,
					(FcChar16 * ) msg, s_len);
	XftFontClose(d, xftfont);
}

/* draw 32bit string into pixmap of sdl_xft_surface  */
void SDL_Xft_Draw_String32(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
						unsigned int x, unsigned int y, FcChar32 *msg) {
	XftFont *xftfont=NULL;
	XRenderColor colortmp;
	XftColor fg;
	XftPattern *patterntmp;

	Display *d;

	int s_len=0;

	d=sdl_xft->d;
	colortmp=xft_surface->color.fg;

	patterntmp=XftPatternDuplicate(xft_surface->fcpattern);
	
	while (msg[s_len] != (FcChar32)0) {
		s_len++;
	}

	xftfont=XftFontOpenPattern (d, patterntmp);

	XftColorAllocValue(d, DefaultVisual(d, 0), 
					DefaultColormap(d,0), &colortmp, &fg);

	XftDrawString32(xft_surface->xftdraw, &fg, xftfont, x, y,
					(FcChar32 * ) msg, s_len);
	XftFontClose(d, xftfont);
}

/* draw utf8 string into pixmap of sdl_xft_surface  */
void SDL_Xft_Draw_StringUtf8(SDL_Xft_WMInfo *sdl_xft, SDL_Xft_Surface *xft_surface, 
						unsigned int x, unsigned int y, FcChar8 *msg) {
	XftFont *xftfont=NULL;
	XRenderColor colortmp;
	XftColor fg;
	XftPattern *patterntmp;

	Display *d;

	int s_len=0;

	d=sdl_xft->d;
	colortmp=xft_surface->color.fg;

	patterntmp=XftPatternDuplicate(xft_surface->fcpattern);
	
	while (msg[s_len] != (FcChar8)0) {
		s_len++;
	}

	xftfont=XftFontOpenPattern (d, patterntmp);

	XftColorAllocValue(d, DefaultVisual(d, 0), 
					DefaultColormap(d,0), &colortmp, &fg);

	XftDrawStringUtf8(xft_surface->xftdraw, &fg, xftfont, x, y,
					(FcChar8 * ) msg, s_len);
	XftFontClose(d, xftfont);
}

/* bliting pixmap into SDL_Surface */
SDL_Surface *SDL_Xft_Blit_Surface(SDL_Xft_WMInfo *sdl_xft, 
								SDL_Xft_Surface *xft_surface) {
	Display *d;
	XImage *ximg;

	XftColor ckey;
	XRenderColor colortmp;

	unsigned int i=0, j=0;
	unsigned long pixel;
	unsigned int key;
	int w=0, h=0;
	Uint32 mpixel;


	d=sdl_xft->d;
	key=xft_surface->color.colorkey;
	switch (key) {
		case SDL_XFT_NO_COLORKEY :
			break;
		case SDL_XFT_BG_COLORKEY :
			colortmp=xft_surface->color.bg;
			break;
		case SDL_XFT_FG_COLORKEY :
			colortmp=xft_surface->color.fg;
			break;
		default :
			break;
	}

	if (key > 0) {
		XftColorAllocValue(d, DefaultVisual(d, 0), 
						DefaultColormap(d,0), &colortmp, &ckey);
	}

	w=xft_surface->surface->w;
	h=xft_surface->surface->h;

	ximg=XGetImage(sdl_xft->d, xft_surface->pixmap, 0, 0, w, h, -1, ZPixmap);

	for ( i=0 ; i < w ; i++) {
		for (j=0; j< h ; j++) {
			pixel=XGetPixel(ximg, i,j);
			if (key > 0  && pixel == ckey.pixel) {
					continue;
			}
			mpixel=GetMapedPixel(xft_surface, pixel);
			putpixel(xft_surface->surface, i, j, mpixel);
			/* for debug
			printf("%#lx\t", pixel);
			*/
		}
	}
	return xft_surface->surface;
}

void SDL_Xft_Quit(SDL_Xft_WMInfo *sdl_xft) {
	FcConfigDestroy(sdl_xft->config);
}


void SDL_Xft_Surface_Free (SDL_Xft_Surface *xft_surface){
	if (xft_surface->fcpattern != NULL) {
		FcPatternDestroy(xft_surface->fcpattern);
	}
	if (xft_surface->xftdraw != NULL) {
		XftDrawDestroy(xft_surface->xftdraw);
	}
	if (xft_surface->extents != NULL) {
		free(xft_surface->extents);
	}
	free(xft_surface);
}

/* private function */
unsigned int FcPattern_Exist(FcPattern *fcpattern) {
	if (fcpattern == NULL) {
		return 0;
	}
	else {
		return 1;
	}
}

/* private function  */
Uint32 GetMapedPixel(SDL_Xft_Surface *xft_surface, unsigned long pixel) {
	Uint32 mpixel;

	SDL_Color color;

	GetSDL_Color(&color, pixel);
	mpixel=SDL_MapRGB (xft_surface->surface->format, color.r, color.g, color.b);

	return mpixel;
}

/* private function  */
void GetSDL_Color(SDL_Color *sdl_color, unsigned long pixel) {
	/* not tested in big endian */
	sdl_color->unused=(Uint8) ((pixel >> 24) & 0xff); /* is alpha channel ? */
	sdl_color->r=(Uint8) ((pixel >> 16) & 0xff);
	sdl_color->g=(Uint8) ((pixel >> 8) & 0xff);
	sdl_color->b=(Uint8) (pixel & 0xff);	
}

/* private function  */
void putpixel(SDL_Surface *surface, unsigned int x, unsigned int y, Uint32 pixel) {
	int bpp = surface->format->BytesPerPixel;
	/* Here p is the address to the pixel we want to set */
	Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

	switch(bpp) {
		case 1:
			*p = pixel;
			break;
			
		case 2:
			*(Uint16 *)p = pixel;
			break;

		case 3:
			if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
				p[0] = (pixel >> 16) & 0xff;
				p[1] = (pixel >> 8) & 0xff;
				p[2] = pixel & 0xff;
			} else {
				p[0] = pixel & 0xff;
				p[1] = (pixel >> 8) & 0xff;
				p[2] = (pixel >> 16) & 0xff;
			}
			break;
		case 4:
			*(Uint32 *)p = pixel;
		break;
	}
}
